# -*- coding: utf-8 -*-
'''
Module to provide Palo Alto compatibility to Salt.

:codeauthor: :email:`Spencer Ervin <spencer_ervin@hotmail.com>`
:maturity:   new
:depends:    none
:platform:   unix

.. versionadded:: 2018.3.0

Configuration
=============

This module accepts connection configuration details either as
parameters, or as configuration settings in pillar as a Salt proxy.
Options passed into opts will be ignored if options are passed into pillar.

.. seealso::
    :prox:`Palo Alto Proxy Module <salt.proxy.panos>`

About
=====

This execution module was designed to handle connections to a Palo Alto based
firewall. This module adds support to send connections directly to the device
through the XML API or through a brokered connection to Panorama.

'''

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals
import logging
import time

# Import Salt Libs
from salt.exceptions import CommandExecutionError
import salt.proxy.panos
import salt.utils.platform

log = logging.getLogger(__name__)

__virtualname__ = 'panos'


def __virtual__():
    '''
    Will load for the panos proxy minions.
    '''
    try:
        if salt.utils.platform.is_proxy() and \
           __opts__['proxy']['proxytype'] == 'panos':
            return __virtualname__
    except KeyError:
        pass

    return False, 'The panos execution module can only be loaded for panos proxy minions.'


def _get_job_results(query=None):
    '''
    Executes a query that requires a job for completion. This function will wait for the job to complete
    and return the results.
    '''
    if not query:
        raise CommandExecutionError("Query parameters cannot be empty.")

    response = __proxy__['panos.call'](query)

    # If the response contains a job, we will wait for the results
    if 'result' in response and 'job' in response['result']:
        jid = response['result']['job']

        while get_job(jid)['result']['job']['status'] != 'FIN':
            time.sleep(5)

        return get_job(jid)
    else:
        return response


def add_config_lock():
    '''
    Prevent other users from changing configuration until the lock is released.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.add_config_lock

    '''
    query = {'type': 'op', 'cmd': '<request><config-lock><add></add></config-lock></request>'}

    return __proxy__['panos.call'](query)


def check_antivirus():
    '''
    Get anti-virus information from PaloAlto Networks server

    CLI Example:

    .. code-block:: bash

        salt '*' panos.check_antivirus

    '''
    query = {'type': 'op', 'cmd': '<request><anti-virus><upgrade><check></check></upgrade></anti-virus></request>'}

    return __proxy__['panos.call'](query)


def check_software():
    '''
    Get software information from PaloAlto Networks server.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.check_software

    '''
    query = {'type': 'op', 'cmd': '<request><system><software><check></check></software></system></request>'}

    return __proxy__['panos.call'](query)


def clear_commit_tasks():
    '''
    Clear all commit tasks.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.clear_commit_tasks

    '''
    query = {'type': 'op', 'cmd': '<request><clear-commit-tasks></clear-commit-tasks></request>'}

    return __proxy__['panos.call'](query)


def commit():
    '''
    Commits the candidate configuration to the running configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.commit

    '''
    query = {'type': 'commit', 'cmd': '<commit></commit>'}

    return _get_job_results(query)


def deactivate_license(key_name=None):
    '''
    Deactivates an installed license.
    Required version 7.0.0 or greater.

    key_name(str): The file name of the license key installed.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.deactivate_license key_name=License_File_Name.key

    '''

    _required_version = '7.0.0'
    if not __proxy__['panos.is_required_version'](_required_version):
        return False, 'The panos device requires version {0} or greater for this command.'.format(_required_version)

    if not key_name:
        return False, 'You must specify a key_name.'
    else:
        query = {'type': 'op', 'cmd': '<request><license><deactivate><key><features><member>{0}</member></features>'
                                      '</key></deactivate></license></request>'.format(key_name)}

    return __proxy__['panos.call'](query)


def delete_license(key_name=None):
    '''
    Remove license keys on disk.

    key_name(str): The file name of the license key to be deleted.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.delete_license key_name=License_File_Name.key

    '''

    if not key_name:
        return False, 'You must specify a key_name.'
    else:
        query = {'type': 'op', 'cmd': '<delete><license><key>{0}</key></license></delete>'.format(key_name)}

    return __proxy__['panos.call'](query)


def download_antivirus():
    '''
    Download the most recent anti-virus package.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.download_antivirus

    '''
    query = {'type': 'op',
             'cmd': '<request><anti-virus><upgrade><download>'
                    '<latest></latest></download></upgrade></anti-virus></request>'}

    return _get_job_results(query)


def download_software_file(filename=None, synch=False):
    '''
    Download software packages by filename.

    Args:
        filename(str): The filename of the PANOS file to download.

        synch (bool): If true then the file will synch to the peer unit.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.download_software_file PanOS_5000-8.0.0
        salt '*' panos.download_software_file PanOS_5000-8.0.0 True

    '''
    if not filename:
        raise CommandExecutionError("Filename option must not be none.")

    if not isinstance(synch, bool):
        raise CommandExecutionError("Synch option must be boolean..")

    if synch is True:
        query = {'type': 'op',
                 'cmd': '<request><system><software><download>'
                        '<file>{0}</file></download></software></system></request>'.format(filename)}
    else:
        query = {'type': 'op',
                 'cmd': '<request><system><software><download><sync-to-peer>yes</sync-to-peer>'
                        '<file>{0}</file></download></software></system></request>'.format(filename)}

    return _get_job_results(query)


def download_software_version(version=None, synch=False):
    '''
    Download software packages by version number.

    Args:
        version(str): The version of the PANOS file to download.

        synch (bool): If true then the file will synch to the peer unit.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.download_software_version 8.0.0
        salt '*' panos.download_software_version 8.0.0 True

    '''
    if not version:
        raise CommandExecutionError("Version option must not be none.")

    if not isinstance(synch, bool):
        raise CommandExecutionError("Synch option must be boolean..")

    if synch is True:
        query = {'type': 'op',
                 'cmd': '<request><system><software><download>'
                        '<version>{0}</version></download></software></system></request>'.format(version)}
    else:
        query = {'type': 'op',
                 'cmd': '<request><system><software><download><sync-to-peer>yes</sync-to-peer>'
                        '<version>{0}</version></download></software></system></request>'.format(version)}

    return _get_job_results(query)


def fetch_license(auth_code=None):
    '''
    Get new license(s) using from the Palo Alto Network Server.

    auth_code
        The license authorization code.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.fetch_license
        salt '*' panos.fetch_license auth_code=foobar

    '''
    if not auth_code:
        query = {'type': 'op', 'cmd': '<request><license><fetch></fetch></license></request>'}
    else:
        query = {'type': 'op', 'cmd': '<request><license><fetch><auth-code>{0}</auth-code></fetch></license>'
                                      '</request>'.format(auth_code)}

    return __proxy__['panos.call'](query)


def get_address(address=None, vsys='1'):
    '''
    Get the candidate configuration for the specified get_address object. This will not return address objects that are
    marked as pre-defined objects.

    address(str): The name of the address object.

    vsys(str): The string representation of the VSYS ID.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_address myhost
        salt '*' panos.get_address myhost 3

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/vsys/entry[@name=\'vsys{0}\']/'
             'address/entry[@name=\'{1}\']'.format(vsys, address)}

    return __proxy__['panos.call'](query)


def get_address_group(addressgroup=None, vsys='1'):
    '''
    Get the candidate configuration for the specified address group. This will not return address groups that are
    marked as pre-defined objects.

    addressgroup(str): The name of the address group.

    vsys(str): The string representation of the VSYS ID.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_address_group foobar
        salt '*' panos.get_address_group foobar 3

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/vsys/entry[@name=\'vsys{0}\']/'
             'address-group/entry[@name=\'{1}\']'.format(vsys, addressgroup)}

    return __proxy__['panos.call'](query)


def get_admins_active():
    '''
    Show active administrators.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_admins_active

    '''
    query = {'type': 'op', 'cmd': '<show><admins></admins></show>'}

    return __proxy__['panos.call'](query)


def get_admins_all():
    '''
    Show all administrators.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_admins_all

    '''
    query = {'type': 'op', 'cmd': '<show><admins><all></all></admins></show>'}

    return __proxy__['panos.call'](query)


def get_antivirus_info():
    '''
    Show information about available anti-virus packages.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_antivirus_info

    '''
    query = {'type': 'op', 'cmd': '<request><anti-virus><upgrade><info></info></upgrade></anti-virus></request>'}

    return __proxy__['panos.call'](query)


def get_arp():
    '''
    Show ARP information.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_arp

    '''
    query = {'type': 'op', 'cmd': '<show><arp><entry name = \'all\'/></arp></show>'}

    return __proxy__['panos.call'](query)


def get_cli_idle_timeout():
    '''
    Show timeout information for this administrative session.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_cli_idle_timeout

    '''
    query = {'type': 'op', 'cmd': '<show><cli><idle-timeout></idle-timeout></cli></show>'}

    return __proxy__['panos.call'](query)


def get_cli_permissions():
    '''
    Show cli administrative permissions.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_cli_permissions

    '''
    query = {'type': 'op', 'cmd': '<show><cli><permissions></permissions></cli></show>'}

    return __proxy__['panos.call'](query)


def get_disk_usage():
    '''
    Report filesystem disk space usage.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_disk_usage

    '''
    query = {'type': 'op', 'cmd': '<show><system><disk-space></disk-space></system></show>'}

    return __proxy__['panos.call'](query)


def get_dns_server_config():
    '''
    Get the DNS server configuration from the candidate configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_dns_server_config

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/dns-setting/servers'}

    return __proxy__['panos.call'](query)


def get_domain_config():
    '''
    Get the domain name configuration from the candidate configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_domain_config

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/domain'}

    return __proxy__['panos.call'](query)


def get_dos_blocks():
    '''
    Show the DoS block-ip table.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_dos_blocks

    '''
    query = {'type': 'op', 'cmd': '<show><dos-block-table><all></all></dos-block-table></show>'}

    return __proxy__['panos.call'](query)


def get_fqdn_cache():
    '''
    Print FQDNs used in rules and their IPs.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_fqdn_cache

    '''
    query = {'type': 'op', 'cmd': '<request><system><fqdn><show></show></fqdn></system></request>'}

    return __proxy__['panos.call'](query)


def get_ha_config():
    '''
    Get the high availability configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_ha_config

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/high-availability'}

    return __proxy__['panos.call'](query)


def get_ha_link():
    '''
    Show high-availability link-monitoring state.

    CLI Example:

   .. code-block:: bash

        salt '*' panos.get_ha_link

    '''
    query = {'type': 'op',
             'cmd': '<show><high-availability><link-monitoring></link-monitoring></high-availability></show>'}

    return __proxy__['panos.call'](query)


def get_ha_path():
    '''
    Show high-availability path-monitoring state.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_ha_path

    '''
    query = {'type': 'op',
             'cmd': '<show><high-availability><path-monitoring></path-monitoring></high-availability></show>'}

    return __proxy__['panos.call'](query)


def get_ha_state():
    '''
    Show high-availability state information.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_ha_state

    '''

    query = {'type': 'op',
             'cmd': '<show><high-availability><state></state></high-availability></show>'}

    return __proxy__['panos.call'](query)


def get_ha_transitions():
    '''
    Show high-availability transition statistic information.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_ha_transitions

    '''

    query = {'type': 'op',
             'cmd': '<show><high-availability><transitions></transitions></high-availability></show>'}

    return __proxy__['panos.call'](query)


def get_hostname():
    '''
    Get the hostname of the device.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_hostname

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/hostname'}

    return __proxy__['panos.call'](query)


def get_interface_counters(name='all'):
    '''
    Get the counter statistics for interfaces.

    Args:
        name (str): The name of the interface to view. By default, all interface statistics are viewed.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_interface_counters
        salt '*' panos.get_interface_counters ethernet1/1

    '''
    query = {'type': 'op',
             'cmd': '<show><counter><interface>{0}</interface></counter></show>'.format(name)}

    return __proxy__['panos.call'](query)


def get_interfaces(name='all'):
    '''
    Show interface information.

    Args:
        name (str): The name of the interface to view. By default, all interface statistics are viewed.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_interfaces
        salt '*' panos.get_interfaces ethernet1/1

    '''
    query = {'type': 'op',
             'cmd': '<show><interface>{0}</interface></show>'.format(name)}

    return __proxy__['panos.call'](query)


def get_job(jid=None):
    '''
    List all a single job by ID.

    jid
        The ID of the job to retrieve.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_job jid=15

    '''
    if not jid:
        raise CommandExecutionError("ID option must not be none.")

    query = {'type': 'op', 'cmd': '<show><jobs><id>{0}</id></jobs></show>'.format(jid)}

    return __proxy__['panos.call'](query)


def get_jobs(state='all'):
    '''
    List all jobs on the device.

    state
        The state of the jobs to display. Valid options are all, pending, or processed. Pending jobs are jobs
        that are currently in a running or waiting state. Processed jobs are jobs that have completed
        execution.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_jobs
        salt '*' panos.get_jobs state=pending

    '''
    if state.lower() == 'all':
        query = {'type': 'op', 'cmd': '<show><jobs><all></all></jobs></show>'}
    elif state.lower() == 'pending':
        query = {'type': 'op', 'cmd': '<show><jobs><pending></pending></jobs></show>'}
    elif state.lower() == 'processed':
        query = {'type': 'op', 'cmd': '<show><jobs><processed></processed></jobs></show>'}
    else:
        raise CommandExecutionError("The state parameter must be all, pending, or processed.")

    return __proxy__['panos.call'](query)


def get_lacp():
    '''
    Show LACP state.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_lacp

    '''
    query = {'type': 'op', 'cmd': '<show><lacp><aggregate-ethernet>all</aggregate-ethernet></lacp></show>'}

    return __proxy__['panos.call'](query)


def get_license_info():
    '''
    Show information about owned license(s).

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_license_info

    '''
    query = {'type': 'op', 'cmd': '<request><license><info></info></license></request>'}

    return __proxy__['panos.call'](query)


def get_license_tokens():
    '''
    Show license token files for manual license deactivation.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_license_tokens

    '''
    query = {'type': 'op', 'cmd': '<show><license-token-files></license-token-files></show>'}

    return __proxy__['panos.call'](query)


def get_lldp_config():
    '''
    Show lldp config for interfaces.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_lldp_config

    '''
    query = {'type': 'op', 'cmd': '<show><lldp><config>all</config></lldp></show>'}

    return __proxy__['panos.call'](query)


def get_lldp_counters():
    '''
    Show lldp counters for interfaces.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_lldp_counters

    '''
    query = {'type': 'op', 'cmd': '<show><lldp><counters>all</counters></lldp></show>'}

    return __proxy__['panos.call'](query)


def get_lldp_local():
    '''
    Show lldp local info for interfaces.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_lldp_local

    '''
    query = {'type': 'op', 'cmd': '<show><lldp><local>all</local></lldp></show>'}

    return __proxy__['panos.call'](query)


def get_lldp_neighbors():
    '''
    Show lldp neighbors info for interfaces.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_lldp_neighbors

    '''
    query = {'type': 'op', 'cmd': '<show><lldp><neighbors>all</neighbors></lldp></show>'}

    return __proxy__['panos.call'](query)


def get_local_admins():
    '''
    Show all local administrator accounts.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_local_admins

    '''
    admin_list = get_users_config()
    response = []

    if 'users' not in admin_list['result']:
        return response

    if isinstance(admin_list['result']['users']['entry'], list):
        for entry in admin_list['result']['users']['entry']:
            response.append(entry['name'])
    else:
        response.append(admin_list['result']['users']['entry']['name'])

    return response


def get_logdb_quota():
    '''
    Report the logdb quotas.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_logdb_quota

    '''
    query = {'type': 'op', 'cmd': '<show><system><logdb-quota></logdb-quota></system></show>'}

    return __proxy__['panos.call'](query)


def get_master_key():
    '''
    Get the master key properties.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_master_key

    '''
    query = {'type': 'op', 'cmd': '<show><system><masterkey-properties></masterkey-properties></system></show>'}

    return __proxy__['panos.call'](query)


def get_ntp_config():
    '''
    Get the NTP configuration from the candidate configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_ntp_config

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/ntp-servers'}

    return __proxy__['panos.call'](query)


def get_ntp_servers():
    '''
    Get list of configured NTP servers.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_ntp_servers

    '''
    query = {'type': 'op', 'cmd': '<show><ntp></ntp></show>'}

    return __proxy__['panos.call'](query)


def get_operational_mode():
    '''
    Show device operational mode setting.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_operational_mode

    '''
    query = {'type': 'op', 'cmd': '<show><operational-mode></operational-mode></show>'}

    return __proxy__['panos.call'](query)


def get_panorama_status():
    '''
    Show panorama connection status.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_panorama_status

    '''
    query = {'type': 'op', 'cmd': '<show><panorama-status></panorama-status></show>'}

    return __proxy__['panos.call'](query)


def get_permitted_ips():
    '''
    Get the IP addresses that are permitted to establish management connections to the device.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_permitted_ips

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/permitted-ip'}

    return __proxy__['panos.call'](query)


def get_platform():
    '''
    Get the platform model information and limitations.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_platform

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/platform'}

    return __proxy__['panos.call'](query)


def get_predefined_application(application=None):
    '''
    Get the configuration for the specified pre-defined application object. This will only return pre-defined
    application objects.

    application(str): The name of the pre-defined application object.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_predefined_application saltstack

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/predefined/application/entry[@name=\'{0}\']'.format(application)}

    return __proxy__['panos.call'](query)


def get_security_rule(rulename=None, vsys='1'):
    '''
    Get the candidate configuration for the specified security rule.

    rulename(str): The name of the security rule.

    vsys(str): The string representation of the VSYS ID.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_security_rule rule01
        salt '*' panos.get_security_rule rule01 3

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/vsys/entry[@name=\'vsys{0}\']/'
             'rulebase/security/rules/entry[@name=\'{1}\']'.format(vsys, rulename)}

    return __proxy__['panos.call'](query)


def get_service(service=None, vsys='1'):
    '''
    Get the candidate configuration for the specified service object. This will not return services that are marked
    as pre-defined objects.

    service(str): The name of the service object.

    vsys(str): The string representation of the VSYS ID.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_service tcp-443
        salt '*' panos.get_service tcp-443 3

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/vsys/entry[@name=\'vsys{0}\']/'
             'service/entry[@name=\'{1}\']'.format(vsys, service)}

    return __proxy__['panos.call'](query)


def get_service_group(servicegroup=None, vsys='1'):
    '''
    Get the candidate configuration for the specified service group. This will not return service groups that are
    marked as pre-defined objects.

    servicegroup(str): The name of the service group.

    vsys(str): The string representation of the VSYS ID.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_service_group foobar
        salt '*' panos.get_service_group foobar 3

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/vsys/entry[@name=\'vsys{0}\']/'
             'service-group/entry[@name=\'{1}\']'.format(vsys, servicegroup)}

    return __proxy__['panos.call'](query)


def get_session_info():
    '''
    Show device session statistics.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_session_info

    '''
    query = {'type': 'op',
             'cmd': '<show><session><info></info></session></show>'}

    return __proxy__['panos.call'](query)


def get_snmp_config():
    '''
    Get the SNMP configuration from the device.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_snmp_config

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/snmp-setting'}

    return __proxy__['panos.call'](query)


def get_software_info():
    '''
    Show information about available software packages.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_software_info

    '''
    query = {'type': 'op', 'cmd': '<request><system><software><info></info></software></system></request>'}

    return __proxy__['panos.call'](query)


def get_system_date_time():
    '''
    Get the system date/time.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_system_date_time

    '''
    query = {'type': 'op', 'cmd': '<show><clock></clock></show>'}

    return __proxy__['panos.call'](query)


def get_system_files():
    '''
    List important files in the system.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_system_files

    '''
    query = {'type': 'op', 'cmd': '<show><system><files></files></system></show>'}

    return __proxy__['panos.call'](query)


def get_system_info():
    '''
    Get the system information.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_system_info

    '''
    query = {'type': 'op', 'cmd': '<show><system><info></info></system></show>'}

    return __proxy__['panos.call'](query)


def get_system_services():
    '''
    Show system services.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_system_services

    '''
    query = {'type': 'op', 'cmd': '<show><system><services></services></system></show>'}

    return __proxy__['panos.call'](query)


def get_system_state(mask=None):
    '''
    Show the system state variables.

    mask
        Filters by a subtree or a wildcard.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_system_state
        salt '*' panos.get_system_state mask=cfg.ha.config.enabled
        salt '*' panos.get_system_state mask=cfg.ha.*

    '''
    if mask:
        query = {'type': 'op',
                 'cmd': '<show><system><state><filter>{0}</filter></state></system></show>'.format(mask)}
    else:
        query = {'type': 'op', 'cmd': '<show><system><state></state></system></show>'}

    return __proxy__['panos.call'](query)


def get_uncommitted_changes():
    '''
    Retrieve a list of all uncommitted changes on the device.
    Requires PANOS version 8.0.0 or greater.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_uncommitted_changes

    '''
    _required_version = '8.0.0'
    if not __proxy__['panos.is_required_version'](_required_version):
        return False, 'The panos device requires version {0} or greater for this command.'.format(_required_version)

    query = {'type': 'op',
             'cmd': '<show><config><list><changes></changes></list></config></show>'}

    return __proxy__['panos.call'](query)


def get_users_config():
    '''
    Get the local administrative user account configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_users_config

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/mgt-config/users'}

    return __proxy__['panos.call'](query)


def get_vlans():
    '''
    Show all VLAN information.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_vlans

    '''
    query = {'type': 'op', 'cmd': '<show><vlan>all</vlan></show>'}

    return __proxy__['panos.call'](query)


def get_xpath(xpath=''):
    '''
    Retrieve a specified xpath from the candidate configuration.

    xpath(str): The specified xpath in the candidate configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_xpath /config/shared/service

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': xpath}

    return __proxy__['panos.call'](query)


def get_zone(zone='', vsys='1'):
    '''
    Get the candidate configuration for the specified zone.

    zone(str): The name of the zone.

    vsys(str): The string representation of the VSYS ID.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_zone trust
        salt '*' panos.get_zone trust 2

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/vsys/entry[@name=\'vsys{0}\']/'
             'zone/entry[@name=\'{1}\']'.format(vsys, zone)}

    return __proxy__['panos.call'](query)


def get_zones(vsys='1'):
    '''
    Get all the zones in the candidate configuration.

    vsys(str): The string representation of the VSYS ID.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.get_zones
        salt '*' panos.get_zones 2

    '''
    query = {'type': 'config',
             'action': 'get',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/vsys/entry[@name=\'vsys{0}\']/'
             'zone'.format(vsys)}

    return __proxy__['panos.call'](query)


def install_antivirus(version=None, latest=False, synch=False, skip_commit=False,):
    '''
    Install anti-virus packages.

    Args:
        version(str): The version of the PANOS file to install.

        latest(bool): If true, the latest anti-virus file will be installed.
                      The specified version option will be ignored.

        synch(bool): If true, the anti-virus will synch to the peer unit.

        skip_commit(bool): If true, the install will skip committing to the device.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.install_antivirus 8.0.0

    '''
    if not version and latest is False:
        raise CommandExecutionError("Version option must not be none.")

    if synch is True:
        s = "yes"
    else:
        s = "no"

    if skip_commit is True:
        c = "yes"
    else:
        c = "no"

    if latest is True:
        query = {'type': 'op',
                 'cmd': '<request><anti-virus><upgrade><install>'
                        '<commit>{0}</commit><sync-to-peer>{1}</sync-to-peer>'
                        '<version>latest</version></install></upgrade></anti-virus></request>'.format(c, s)}
    else:
        query = {'type': 'op',
                 'cmd': '<request><anti-virus><upgrade><install>'
                        '<commit>{0}</commit><sync-to-peer>{1}</sync-to-peer>'
                        '<version>{2}</version></install></upgrade></anti-virus></request>'.format(c, s, version)}

    return _get_job_results(query)


def install_license():
    '''
    Install the license key(s).

    CLI Example:

    .. code-block:: bash

        salt '*' panos.install_license

    '''
    query = {'type': 'op', 'cmd': '<request><license><install></install></license></request>'}

    return __proxy__['panos.call'](query)


def install_software(version=None):
    '''
    Upgrade to a software package by version.

    Args:
        version(str): The version of the PANOS file to install.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.install_license 8.0.0

    '''
    if not version:
        raise CommandExecutionError("Version option must not be none.")

    query = {'type': 'op',
             'cmd': '<request><system><software><install>'
                    '<version>{0}</version></install></software></system></request>'.format(version)}

    return _get_job_results(query)


def reboot():
    '''
    Reboot a running system.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.reboot

    '''
    query = {'type': 'op', 'cmd': '<request><restart><system></system></restart></request>'}

    return __proxy__['panos.call'](query)


def refresh_fqdn_cache(force=False):
    '''
    Force refreshes all FQDNs used in rules.

    force
        Forces all fqdn refresh

    CLI Example:

    .. code-block:: bash

        salt '*' panos.refresh_fqdn_cache
        salt '*' panos.refresh_fqdn_cache force=True

    '''
    if not isinstance(force, bool):
        raise CommandExecutionError("Force option must be boolean.")

    if force:
        query = {'type': 'op',
                 'cmd': '<request><system><fqdn><refresh><force>yes</force></refresh></fqdn></system></request>'}
    else:
        query = {'type': 'op', 'cmd': '<request><system><fqdn><refresh></refresh></fqdn></system></request>'}

    return __proxy__['panos.call'](query)


def remove_config_lock():
    '''
    Release config lock previously held.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.remove_config_lock

    '''
    query = {'type': 'op', 'cmd': '<request><config-lock><remove></remove></config-lock></request>'}

    return __proxy__['panos.call'](query)


def resolve_address(address=None, vsys=None):
    '''
    Resolve address to ip address.
    Required version 7.0.0 or greater.

    address
        Address name you want to resolve.

    vsys
        The vsys name.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.resolve_address foo.bar.com
        salt '*' panos.resolve_address foo.bar.com vsys=2

    '''
    _required_version = '7.0.0'
    if not __proxy__['panos.is_required_version'](_required_version):
        return False, 'The panos device requires version {0} or greater for this command.'.format(_required_version)

    if not address:
        raise CommandExecutionError("FQDN to resolve must be provided as address.")

    if not vsys:
        query = {'type': 'op',
                 'cmd': '<request><resolve><address>{0}</address></resolve></request>'.format(address)}
    else:
        query = {'type': 'op',
                 'cmd': '<request><resolve><vsys>{0}</vsys><address>{1}</address></resolve>'
                        '</request>'.format(vsys, address)}

    return __proxy__['panos.call'](query)


def save_device_config(filename=None):
    '''
    Save device configuration to a named file.

    filename
        The filename to save the configuration to.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.save_device_config foo.xml

    '''
    if not filename:
        raise CommandExecutionError("Filename must not be empty.")

    query = {'type': 'op', 'cmd': '<save><config><to>{0}</to></config></save>'.format(filename)}

    return __proxy__['panos.call'](query)


def save_device_state():
    '''
    Save files needed to restore device to local disk.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.save_device_state

    '''
    query = {'type': 'op', 'cmd': '<save><device-state></device-state></save>'}

    return __proxy__['panos.call'](query)


def set_authentication_profile(profile=None, deploy=False):
    '''
    Set the authentication profile of the Palo Alto proxy minion. A commit will be required before this is processed.

    CLI Example:

    Args:
        profile (str): The name of the authentication profile to set.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_authentication_profile foo
        salt '*' panos.set_authentication_profile foo deploy=True

    '''

    if not profile:
        CommandExecutionError("Profile name option must not be none.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/'
                      'authentication-profile',
             'element': '<authentication-profile>{0}</authentication-profile>'.format(profile)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_hostname(hostname=None, deploy=False):
    '''
    Set the hostname of the Palo Alto proxy minion. A commit will be required before this is processed.

    CLI Example:

    Args:
        hostname (str): The hostname to set

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_hostname newhostname
        salt '*' panos.set_hostname newhostname deploy=True

    '''

    if not hostname:
        CommandExecutionError("Hostname option must not be none.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system',
             'element': '<hostname>{0}</hostname>'.format(hostname)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_management_icmp(enabled=True, deploy=False):
    '''
    Enables or disables the ICMP management service on the device.

    CLI Example:

    Args:
        enabled (bool): If true the service will be enabled. If false the service will be disabled.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_management_icmp
        salt '*' panos.set_management_icmp enabled=False deploy=True

    '''

    if enabled is True:
        value = "no"
    elif enabled is False:
        value = "yes"
    else:
        CommandExecutionError("Invalid option provided for service enabled option.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/service',
             'element': '<disable-icmp>{0}</disable-icmp>'.format(value)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_management_http(enabled=True, deploy=False):
    '''
    Enables or disables the HTTP management service on the device.

    CLI Example:

    Args:
        enabled (bool): If true the service will be enabled. If false the service will be disabled.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_management_http
        salt '*' panos.set_management_http enabled=False deploy=True

    '''

    if enabled is True:
        value = "no"
    elif enabled is False:
        value = "yes"
    else:
        CommandExecutionError("Invalid option provided for service enabled option.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/service',
             'element': '<disable-http>{0}</disable-http>'.format(value)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_management_https(enabled=True, deploy=False):
    '''
    Enables or disables the HTTPS management service on the device.

    CLI Example:

    Args:
        enabled (bool): If true the service will be enabled. If false the service will be disabled.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_management_https
        salt '*' panos.set_management_https enabled=False deploy=True

    '''

    if enabled is True:
        value = "no"
    elif enabled is False:
        value = "yes"
    else:
        CommandExecutionError("Invalid option provided for service enabled option.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/service',
             'element': '<disable-https>{0}</disable-https>'.format(value)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_management_ocsp(enabled=True, deploy=False):
    '''
    Enables or disables the HTTP OCSP management service on the device.

    CLI Example:

    Args:
        enabled (bool): If true the service will be enabled. If false the service will be disabled.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_management_ocsp
        salt '*' panos.set_management_ocsp enabled=False deploy=True

    '''

    if enabled is True:
        value = "no"
    elif enabled is False:
        value = "yes"
    else:
        CommandExecutionError("Invalid option provided for service enabled option.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/service',
             'element': '<disable-http-ocsp>{0}</disable-http-ocsp>'.format(value)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_management_snmp(enabled=True, deploy=False):
    '''
    Enables or disables the SNMP management service on the device.

    CLI Example:

    Args:
        enabled (bool): If true the service will be enabled. If false the service will be disabled.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_management_snmp
        salt '*' panos.set_management_snmp enabled=False deploy=True

    '''

    if enabled is True:
        value = "no"
    elif enabled is False:
        value = "yes"
    else:
        CommandExecutionError("Invalid option provided for service enabled option.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/service',
             'element': '<disable-snmp>{0}</disable-snmp>'.format(value)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_management_ssh(enabled=True, deploy=False):
    '''
    Enables or disables the SSH management service on the device.

    CLI Example:

    Args:
        enabled (bool): If true the service will be enabled. If false the service will be disabled.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_management_ssh
        salt '*' panos.set_management_ssh enabled=False deploy=True

    '''

    if enabled is True:
        value = "no"
    elif enabled is False:
        value = "yes"
    else:
        CommandExecutionError("Invalid option provided for service enabled option.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/service',
             'element': '<disable-ssh>{0}</disable-ssh>'.format(value)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_management_telnet(enabled=True, deploy=False):
    '''
    Enables or disables the Telnet management service on the device.

    CLI Example:

    Args:
        enabled (bool): If true the service will be enabled. If false the service will be disabled.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_management_telnet
        salt '*' panos.set_management_telnet enabled=False deploy=True

    '''

    if enabled is True:
        value = "no"
    elif enabled is False:
        value = "yes"
    else:
        CommandExecutionError("Invalid option provided for service enabled option.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/service',
             'element': '<disable-telnet>{0}</disable-telnet>'.format(value)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_ntp_authentication(target=None,
                           authentication_type=None,
                           key_id=None,
                           authentication_key=None,
                           algorithm=None,
                           deploy=False):
    '''
    Set the NTP authentication of the Palo Alto proxy minion. A commit will be required before this is processed.

    CLI Example:

    Args:
        target(str): Determines the target of the authentication. Valid options are primary, secondary, or both.

        authentication_type(str): The authentication type to be used. Valid options are symmetric, autokey, and none.

        key_id(int): The NTP authentication key ID.

        authentication_key(str): The authentication key.

        algorithm(str): The algorithm type to be used for a symmetric key. Valid options are md5 and sha1.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' ntp.set_authentication target=both authentication_type=autokey
        salt '*' ntp.set_authentication target=primary authentication_type=none
        salt '*' ntp.set_authentication target=both authentication_type=symmetric key_id=15 authentication_key=mykey algorithm=md5
        salt '*' ntp.set_authentication target=both authentication_type=symmetric key_id=15 authentication_key=mykey algorithm=md5 deploy=True

    '''
    ret = {}

    if target not in ['primary', 'secondary', 'both']:
        raise salt.exceptions.CommandExecutionError("Target option must be primary, secondary, or both.")

    if authentication_type not in ['symmetric', 'autokey', 'none']:
        raise salt.exceptions.CommandExecutionError("Type option must be symmetric, autokey, or both.")

    if authentication_type == "symmetric" and not authentication_key:
        raise salt.exceptions.CommandExecutionError("When using symmetric authentication, authentication_key must be "
                                                    "provided.")

    if authentication_type == "symmetric" and not key_id:
        raise salt.exceptions.CommandExecutionError("When using symmetric authentication, key_id must be provided.")

    if authentication_type == "symmetric" and algorithm not in ['md5', 'sha1']:
        raise salt.exceptions.CommandExecutionError("When using symmetric authentication, algorithm must be md5 or "
                                                    "sha1.")

    if authentication_type == 'symmetric':
        if target == 'primary' or target == 'both':
            query = {'type': 'config',
                     'action': 'set',
                     'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/ntp-servers/'
                              'primary-ntp-server/authentication-type',
                     'element': '<symmetric-key><algorithm><{0}><authentication-key>{1}</authentication-key></{0}>'
                                '</algorithm><key-id>{2}</key-id></symmetric-key>'.format(algorithm,
                                                                                          authentication_key,
                                                                                          key_id)}
            ret.update({'primary_server': __proxy__['panos.call'](query)})

        if target == 'secondary' or target == 'both':
            query = {'type': 'config',
                     'action': 'set',
                     'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/ntp-servers/'
                              'secondary-ntp-server/authentication-type',
                     'element': '<symmetric-key><algorithm><{0}><authentication-key>{1}</authentication-key></{0}>'
                                '</algorithm><key-id>{2}</key-id></symmetric-key>'.format(algorithm,
                                                                                          authentication_key,
                                                                                          key_id)}
            ret.update({'secondary_server': __proxy__['panos.call'](query)})
    elif authentication_type == 'autokey':
        if target == 'primary' or target == 'both':
            query = {'type': 'config',
                     'action': 'set',
                     'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/ntp-servers/'
                              'primary-ntp-server/authentication-type',
                     'element': '<autokey/>'}
            ret.update({'primary_server': __proxy__['panos.call'](query)})

        if target == 'secondary' or target == 'both':
            query = {'type': 'config',
                     'action': 'set',
                     'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/ntp-servers/'
                              'secondary-ntp-server/authentication-type',
                     'element': '<autokey/>'}
            ret.update({'secondary_server': __proxy__['panos.call'](query)})
    elif authentication_type == 'none':
        if target == 'primary' or target == 'both':
            query = {'type': 'config',
                     'action': 'set',
                     'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/ntp-servers/'
                              'primary-ntp-server/authentication-type',
                     'element': '<none/>'}
            ret.update({'primary_server': __proxy__['panos.call'](query)})

        if target == 'secondary' or target == 'both':
            query = {'type': 'config',
                     'action': 'set',
                     'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/ntp-servers/'
                              'secondary-ntp-server/authentication-type',
                     'element': '<none/>'}
            ret.update({'secondary_server': __proxy__['panos.call'](query)})

    if deploy is True:
        ret.update(commit())

    return ret


def set_ntp_servers(primary_server=None, secondary_server=None, deploy=False):
    '''
    Set the NTP servers of the Palo Alto proxy minion. A commit will be required before this is processed.

    CLI Example:

    Args:
        primary_server(str): The primary NTP server IP address or FQDN.

        secondary_server(str): The secondary NTP server IP address or FQDN.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' ntp.set_servers 0.pool.ntp.org 1.pool.ntp.org
        salt '*' ntp.set_servers primary_server=0.pool.ntp.org secondary_server=1.pool.ntp.org
        salt '*' ntp.ser_servers 0.pool.ntp.org 1.pool.ntp.org deploy=True

    '''
    ret = {}

    if primary_server:
        query = {'type': 'config',
                 'action': 'set',
                 'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/ntp-servers/'
                          'primary-ntp-server',
                 'element': '<ntp-server-address>{0}</ntp-server-address>'.format(primary_server)}
        ret.update({'primary_server': __proxy__['panos.call'](query)})

    if secondary_server:
        query = {'type': 'config',
                 'action': 'set',
                 'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/ntp-servers/'
                          'secondary-ntp-server',
                 'element': '<ntp-server-address>{0}</ntp-server-address>'.format(secondary_server)}
        ret.update({'secondary_server': __proxy__['panos.call'](query)})

    if deploy is True:
        ret.update(commit())

    return ret


def set_permitted_ip(address=None, deploy=False):
    '''
    Add an IPv4 address or network to the permitted IP list.

    CLI Example:

    Args:
        address (str): The IPv4 address or network to allow access to add to the Palo Alto device.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_permitted_ip 10.0.0.1
        salt '*' panos.set_permitted_ip 10.0.0.0/24
        salt '*' panos.set_permitted_ip 10.0.0.1 deploy=True

    '''

    if not address:
        CommandExecutionError("Address option must not be empty.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/permitted-ip',
             'element': '<entry name=\'{0}\'></entry>'.format(address)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def set_timezone(tz=None, deploy=False):
    '''
    Set the timezone of the Palo Alto proxy minion. A commit will be required before this is processed.

    CLI Example:

    Args:
        tz (str): The name of the timezone to set.

        deploy (bool): If true then commit the full candidate configuration, if false only set pending change.

    .. code-block:: bash

        salt '*' panos.set_timezone UTC
        salt '*' panos.set_timezone UTC deploy=True

    '''

    if not tz:
        CommandExecutionError("Timezone name option must not be none.")

    ret = {}

    query = {'type': 'config',
             'action': 'set',
             'xpath': '/config/devices/entry[@name=\'localhost.localdomain\']/deviceconfig/system/timezone',
             'element': '<timezone>{0}</timezone>'.format(tz)}

    ret.update(__proxy__['panos.call'](query))

    if deploy is True:
        ret.update(commit())

    return ret


def shutdown():
    '''
    Shutdown a running system.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.shutdown

    '''
    query = {'type': 'op', 'cmd': '<request><shutdown><system></system></shutdown></request>'}

    return __proxy__['panos.call'](query)


def test_fib_route(ip=None,
                   vr='vr1'):
    '''
    Perform a route lookup within active route table (fib).

    ip (str): The destination IP address to test.

    vr (str): The name of the virtual router to test.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.test_fib_route 4.2.2.2
        salt '*' panos.test_fib_route 4.2.2.2 my-vr

    '''

    xpath = "<test><routing><fib-lookup>"

    if ip:
        xpath += "<ip>{0}</ip>".format(ip)

    if vr:
        xpath += "<virtual-router>{0}</virtual-router>".format(vr)

    xpath += "</fib-lookup></routing></test>"

    query = {'type': 'op',
             'cmd': xpath}

    return __proxy__['panos.call'](query)


def test_security_policy(sourcezone=None,
                         destinationzone=None,
                         source=None,
                         destination=None,
                         protocol=None,
                         port=None,
                         application=None,
                         category=None,
                         vsys='1',
                         allrules=False):
    '''
    Checks which security policy as connection will match on the device.

    sourcezone (str): The source zone matched against the connection.

    destinationzone (str): The destination zone matched against the connection.

    source (str): The source address. This must be a single IP address.

    destination (str): The destination address. This must be a single IP address.

    protocol (int): The protocol number for the connection. This is the numerical representation of the protocol.

    port (int): The port number for the connection.

    application (str): The application that should be matched.

    category (str): The category that should be matched.

    vsys (int): The numerical representation of the VSYS ID.

    allrules (bool): Show all potential match rules until first allow rule.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.test_security_policy sourcezone=trust destinationzone=untrust protocol=6 port=22
        salt '*' panos.test_security_policy sourcezone=trust destinationzone=untrust protocol=6 port=22 vsys=2

    '''

    xpath = "<test><security-policy-match>"

    if sourcezone:
        xpath += "<from>{0}</from>".format(sourcezone)

    if destinationzone:
        xpath += "<to>{0}</to>".format(destinationzone)

    if source:
        xpath += "<source>{0}</source>".format(source)

    if destination:
        xpath += "<destination>{0}</destination>".format(destination)

    if protocol:
        xpath += "<protocol>{0}</protocol>".format(protocol)

    if port:
        xpath += "<destination-port>{0}</destination-port>".format(port)

    if application:
        xpath += "<application>{0}</application>".format(application)

    if category:
        xpath += "<category>{0}</category>".format(category)

    if allrules:
        xpath += "<show-all>yes</show-all>"

    xpath += "</security-policy-match></test>"

    query = {'type': 'op',
             'vsys': "vsys{0}".format(vsys),
             'cmd': xpath}

    return __proxy__['panos.call'](query)


def unlock_admin(username=None):
    '''
    Unlocks a locked administrator account.

    username
        Username of the administrator.

    CLI Example:

    .. code-block:: bash

        salt '*' panos.unlock_admin username=bob

    '''
    if not username:
        raise CommandExecutionError("Username option must not be none.")

    query = {'type': 'op',
             'cmd': '<set><management-server><unlock><admin>{0}</admin></unlock></management-server>'
                    '</set>'.format(username)}

    return __proxy__['panos.call'](query)



# -*- coding: utf-8 -*-
'''
Management of user groups
=========================

The group module is used to create and manage group settings, groups can be
either present or absent. User/Group names can be passed to the ``adduser``,
``deluser``, and ``members`` parameters. ``adduser`` and ``deluser`` can be used
together but not with ``members``.

In Windows, if no domain is specified in the user or group name (i.e.
``DOMAIN\\username``) the module will assume a local user or group.

.. code-block:: yaml

    cheese:
      group.present:
        - gid: 7648
        - system: True
        - addusers:
          - user1
          - users2
        - delusers:
          - foo

    cheese:
      group.present:
        - gid: 7648
        - system: True
        - members:
          - foo
          - bar
          - user1
          - user2
'''

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals
import sys

# Import 3rd-party libs
from salt.ext import six

# Import Salt libs
import salt.utils.platform
import salt.utils.win_functions


def _changes(name,
             gid=None,
             addusers=None,
             delusers=None,
             members=None):
    '''
    Return a dict of the changes required for a group if the group is present,
    otherwise return False.
    '''
    lgrp = __salt__['group.info'](name)
    if not lgrp:
        return False

    # User and Domain names are not case sensitive in Windows. Let's make them
    # all lower case so we can compare properly
    if salt.utils.platform.is_windows():
        if lgrp['members']:
            lgrp['members'] = [user.lower() for user in lgrp['members']]
        if members:
            members = [salt.utils.win_functions.get_sam_name(user).lower() for user in members]
        if addusers:
            addusers = [salt.utils.win_functions.get_sam_name(user).lower() for user in addusers]
        if delusers:
            delusers = [salt.utils.win_functions.get_sam_name(user).lower() for user in delusers]

    change = {}
    ret = {}
    if gid:
        try:
            gid = int(gid)
            if lgrp['gid'] != gid:
                change['gid'] = gid
        except (TypeError, ValueError):
            ret['result'] = False
            ret['comment'] = 'Invalid gid'
            return ret

    if members:
        # -- if new member list if different than the current
        if set(lgrp['members']).symmetric_difference(members):
            change['members'] = members

    if addusers:
        users_2add = [user for user in addusers if user not in lgrp['members']]
        if users_2add:
            change['addusers'] = users_2add

    if delusers:
        users_2del = [user for user in delusers if user in lgrp['members']]
        if users_2del:
            change['delusers'] = users_2del

    return change


def present(name,
            gid=None,
            system=False,
            addusers=None,
            delusers=None,
            members=None):
    r'''
    Ensure that a group is present

    Args:

        name (str):
            The name of the group to manage

        gid (str):
            The group id to assign to the named group; if left empty, then the
            next available group id will be assigned. Ignored on Windows

        system (bool):
            Whether or not the named group is a system group.  This is essentially
            the '-r' option of 'groupadd'. Ignored on Windows

        addusers (list):
            List of additional users to be added as a group members. Cannot
            conflict with names in delusers. Cannot be used in conjunction with
            members.

        delusers (list):
            Ensure these user are removed from the group membership. Cannot
            conflict with names in addusers. Cannot be used in conjunction with
            members.

        members (list):
            Replace existing group members with a list of new members. Cannot be
            used in conjunction with addusers or delusers.

    Example:

    .. code-block:: yaml

        # Adds DOMAIN\db_admins and Administrators to the local db_admin group
        # Removes Users
        db_admin:
          group.present:
            - addusers:
              - DOMAIN\db_admins
              - Administrators
            - delusers:
              - Users

        # Ensures only DOMAIN\domain_admins and the local Administrator are
        # members of the local Administrators group. All other users are
        # removed
        Administrators:
          group.present:
            - members:
              - DOMAIN\domain_admins
              - Administrator
    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': 'Group {0} is present and up to date'.format(name)}

    if members and (addusers or delusers):
        ret['result'] = None
        ret['comment'] = (
            'Error: Conflicting options "members" with "addusers" and/or'
            ' "delusers" can not be used together. ')
        return ret

    if addusers and delusers:
        # -- if trying to add and delete the same user(s) at the same time.
        if not set(addusers).isdisjoint(set(delusers)):
            ret['result'] = None
            ret['comment'] = (
                'Error. Same user(s) can not be added and deleted'
                ' simultaneously')
            return ret

    changes = _changes(name,
                       gid,
                       addusers,
                       delusers,
                       members)
    if changes:
        ret['comment'] = (
            'The following group attributes are set to be changed:\n')
        for key, val in six.iteritems(changes):
            ret['comment'] += '{0}: {1}\n'.format(key, val)

        if __opts__['test']:
            ret['result'] = None
            return ret

        for key, val in six.iteritems(changes):
            if key == 'gid':
                __salt__['group.chgid'](name, gid)
                continue
            if key == 'addusers':
                for user in val:
                    __salt__['group.adduser'](name, user)
                continue
            if key == 'delusers':
                for user in val:
                    __salt__['group.deluser'](name, user)
                continue
            if key == 'members':
                __salt__['group.members'](name, ','.join(members))
                continue
        # Clear cached group data
        sys.modules[
            __salt__['test.ping'].__module__
            ].__context__.pop('group.getent', None)
        changes = _changes(name,
                           gid,
                           addusers,
                           delusers,
                           members)
        if changes:
            ret['result'] = False
            ret['comment'] += 'Some changes could not be applied'
            ret['changes'] = {'Failed': changes}
        else:
            ret['changes'] = {'Final': 'All changes applied successfully'}

    if changes is False:
        # The group is not present, make it!
        if __opts__['test']:
            ret['result'] = None
            ret['comment'] = 'Group {0} set to be added'.format(name)
            return ret

        grps = __salt__['group.getent']()
        # Test if gid is free
        if gid is not None:
            gid_group = None
            for lgrp in grps:
                if lgrp['gid'] == gid:
                    gid_group = lgrp['name']
                    break

            if gid_group is not None:
                ret['result'] = False
                ret['comment'] = (
                    'Group {0} is not present but gid {1} is already taken by'
                    ' group {2}'.format(name, gid, gid_group))
                return ret

        # Group is not present, make it.
        if __salt__['group.add'](name, gid=gid, system=system):
            # if members to be added
            grp_members = None
            if members:
                grp_members = ','.join(members)
            if addusers:
                grp_members = ','.join(addusers)
            if grp_members:
                __salt__['group.members'](name, grp_members)
            # Clear cached group data
            sys.modules[__salt__['test.ping'].__module__].__context__.pop(
                'group.getent', None)
            ret['comment'] = 'New group {0} created'.format(name)
            ret['changes'] = __salt__['group.info'](name)
            changes = _changes(name,
                               gid,
                               addusers,
                               delusers,
                               members)
            if changes:
                ret['result'] = False
                ret['comment'] = (
                    'Group {0} has been created but, some changes could not'
                    ' be applied'.format(name))
                ret['changes'] = {'Failed': changes}
        else:
            ret['result'] = False
            ret['comment'] = 'Failed to create new group {0}'.format(name)
    return ret


def absent(name):
    '''
    Ensure that the named group is absent

    Args:
        name (str):
            The name of the group to remove

    Example:

    .. code-block:: yaml

        # Removes the local group `db_admin`
        db_admin:
          group.absent
    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': ''}
    grp_info = __salt__['group.info'](name)
    if grp_info:
        # Group already exists. Remove the group.
        if __opts__['test']:
            ret['result'] = None
            ret['comment'] = 'Group {0} is set for removal'.format(name)
            return ret
        ret['result'] = __salt__['group.delete'](name)
        if ret['result']:
            ret['changes'] = {name: ''}
            ret['comment'] = 'Removed group {0}'.format(name)
            return ret
        else:
            ret['comment'] = 'Failed to remove group {0}'.format(name)
            return ret
    else:
        ret['comment'] = 'Group not present'
        return ret



# -*- coding: utf-8 -*-
'''
Connection library for Amazon S3

:depends: requests
'''
from __future__ import absolute_import, print_function, unicode_literals

# Import Python libs
import logging

# Import 3rd-party libs
try:
    import requests
    HAS_REQUESTS = True  # pylint: disable=W0612
except ImportError:
    HAS_REQUESTS = False  # pylint: disable=W0612

# Import Salt libs
import salt.utils.aws
import salt.utils.files
import salt.utils.hashutils
import salt.utils.xmlutil as xml
from salt._compat import ElementTree as ET
from salt.exceptions import CommandExecutionError
from salt.ext.six.moves.urllib.parse import quote as _quote   # pylint: disable=import-error,no-name-in-module
from salt.ext import six

log = logging.getLogger(__name__)


def query(key, keyid, method='GET', params=None, headers=None,
          requesturl=None, return_url=False, bucket=None, service_url=None,
          path='', return_bin=False, action=None, local_file=None,
          verify_ssl=True, full_headers=False, kms_keyid=None,
          location=None, role_arn=None, chunk_size=16384, path_style=False,
          https_enable=True):
    '''
    Perform a query against an S3-like API. This function requires that a
    secret key and the id for that key are passed in. For instance:

        s3.keyid: GKTADJGHEIQSXMKKRBJ08H
        s3.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

    If keyid or key is not specified, an attempt to fetch them from EC2 IAM
    metadata service will be made.

    A service_url may also be specified in the configuration:

        s3.service_url: s3.amazonaws.com

    If a service_url is not specified, the default is s3.amazonaws.com. This
    may appear in various documentation as an "endpoint". A comprehensive list
    for Amazon S3 may be found at::

        http://docs.aws.amazon.com/general/latest/gr/rande.html#s3_region

    The service_url will form the basis for the final endpoint that is used to
    query the service.

    Path style can be enabled:

        s3.path_style: True

    This can be useful if you need to use salt with a proxy for an s3 compatible storage

    You can use either https protocol or http protocol:

        s3.https_enable: True

    SSL verification may also be turned off in the configuration:

        s3.verify_ssl: False

    This is required if using S3 bucket names that contain a period, as
    these will not match Amazon's S3 wildcard certificates. Certificate
    verification is enabled by default.

    A region may be specified:

        s3.location: eu-central-1

    If region is not specified, an attempt to fetch the region from EC2 IAM
    metadata service will be made. Failing that, default is us-east-1
    '''
    if not HAS_REQUESTS:
        log.error('There was an error: requests is required for s3 access')

    if not headers:
        headers = {}

    if not params:
        params = {}

    if not service_url:
        service_url = 's3.amazonaws.com'

    if not bucket or path_style:
        endpoint = service_url
    else:
        endpoint = '{0}.{1}'.format(bucket, service_url)

    if path_style and bucket:
        path = '{0}/{1}'.format(bucket, path)

    # Try grabbing the credentials from the EC2 instance IAM metadata if available
    if not key:
        key = salt.utils.aws.IROLE_CODE

    if not keyid:
        keyid = salt.utils.aws.IROLE_CODE

    if kms_keyid is not None and method in ('PUT', 'POST'):
        headers['x-amz-server-side-encryption'] = 'aws:kms'
        headers['x-amz-server-side-encryption-aws-kms-key-id'] = kms_keyid

    if not location:
        location = salt.utils.aws.get_location()

    data = ''
    fh = None
    payload_hash = None
    if method == 'PUT':
        if local_file:
            payload_hash = salt.utils.hashutils.get_hash(local_file, form='sha256')

    if path is None:
        path = ''
    path = _quote(path)

    if not requesturl:
        requesturl = (('https' if https_enable else 'http')+'://{0}/{1}').format(endpoint, path)
        headers, requesturl = salt.utils.aws.sig4(
            method,
            endpoint,
            params,
            data=data,
            uri='/{0}'.format(path),
            prov_dict={'id': keyid, 'key': key},
            role_arn=role_arn,
            location=location,
            product='s3',
            requesturl=requesturl,
            headers=headers,
            payload_hash=payload_hash,
        )

    log.debug('S3 Request: %s', requesturl)
    log.debug('S3 Headers::')
    log.debug('    Authorization: %s', headers['Authorization'])

    if not data:
        data = None

    try:
        if method == 'PUT':
            if local_file:
                fh = salt.utils.files.fopen(local_file, 'rb')  # pylint: disable=resource-leakage
                data = fh.read()  # pylint: disable=resource-leakage
            result = requests.request(method,
                                      requesturl,
                                      headers=headers,
                                      data=data,
                                      verify=verify_ssl,
                                      stream=True,
                                      timeout=300)
        elif method == 'GET' and local_file and not return_bin:
            result = requests.request(method,
                                      requesturl,
                                      headers=headers,
                                      data=data,
                                      verify=verify_ssl,
                                      stream=True,
                                      timeout=300)
        else:
            result = requests.request(method,
                                      requesturl,
                                      headers=headers,
                                      data=data,
                                      verify=verify_ssl,
                                      timeout=300)
    finally:
        if fh is not None:
            fh.close()

    err_code = None
    err_msg = None
    if result.status_code >= 400:
        # On error the S3 API response should contain error message
        err_text = result.content or 'Unknown error'
        log.debug('    Response content: %s', err_text)

        # Try to get err info from response xml
        try:
            err_data = xml.to_dict(ET.fromstring(err_text))
            err_code = err_data['Code']
            err_msg = err_data['Message']
        except (KeyError, ET.ParseError) as err:
            log.debug(
                'Failed to parse s3 err response. %s: %s',
                type(err).__name__, err
            )
            err_code = 'http-{0}'.format(result.status_code)
            err_msg = err_text

    log.debug('S3 Response Status Code: %s', result.status_code)

    if method == 'PUT':
        if result.status_code != 200:
            if local_file:
                raise CommandExecutionError(
                    'Failed to upload from {0} to {1}. {2}: {3}'.format(
                        local_file, path, err_code, err_msg))
            raise CommandExecutionError(
                'Failed to create bucket {0}. {1}: {2}'.format(
                    bucket, err_code, err_msg))

        if local_file:
            log.debug('Uploaded from %s to %s', local_file, path)
        else:
            log.debug('Created bucket %s', bucket)
        return

    if method == 'DELETE':
        if not six.text_type(result.status_code).startswith('2'):
            if path:
                raise CommandExecutionError(
                    'Failed to delete {0} from bucket {1}. {2}: {3}'.format(
                        path, bucket, err_code, err_msg))
            raise CommandExecutionError(
                'Failed to delete bucket {0}. {1}: {2}'.format(
                    bucket, err_code, err_msg))

        if path:
            log.debug('Deleted %s from bucket %s', path, bucket)
        else:
            log.debug('Deleted bucket %s', bucket)
        return

    # This can be used to save a binary object to disk
    if local_file and method == 'GET':
        if result.status_code < 200 or result.status_code >= 300:
            raise CommandExecutionError(
                'Failed to get file. {0}: {1}'.format(err_code, err_msg))

        log.debug('Saving to local file: %s', local_file)
        with salt.utils.files.fopen(local_file, 'wb') as out:
            for chunk in result.iter_content(chunk_size=chunk_size):
                out.write(chunk)
        return 'Saved to local file: {0}'.format(local_file)

    if result.status_code < 200 or result.status_code >= 300:
        raise CommandExecutionError(
            'Failed s3 operation. {0}: {1}'.format(err_code, err_msg))

    # This can be used to return a binary object wholesale
    if return_bin:
        return result.content

    if result.content:
        items = ET.fromstring(result.content)

        ret = []
        for item in items:
            ret.append(xml.to_dict(item))

        if return_url is True:
            return ret, requesturl
    else:
        if result.status_code != requests.codes.ok:
            return
        ret = {'headers': []}
        if full_headers:
            ret['headers'] = dict(result.headers)
        else:
            for header in result.headers:
                ret['headers'].append(header.strip())

    return ret



# -*- coding: utf-8 -*-
'''
Allows you to manage extended attributes on files or directories
================================================================

Install, enable and disable assistive access on macOS minions

.. code-block:: yaml

    /path/to/file:
      xattr.exists:
        - attributes:
            - com.file.attr=test
            - com.apple.quarantine=0x00001111
'''
from __future__ import absolute_import, unicode_literals, print_function

# Import python libs
import logging
import os

log = logging.getLogger(__name__)
__virtualname__ = "xattr"


def __virtual__():
    '''
    Only work on Mac OS
    '''
    if __grains__['os'] in ['MacOS', 'Darwin']:
        return __virtualname__
    return False


def exists(name, attributes):
    '''
    Make sure the given attributes exist on the file/directory

    name
        The path to the file/directory

    attributes
        The attributes that should exist on the file/directory, this is accepted as
        an array, with key and value split with an equals sign, if you want to specify
        a hex value then add 0x to the beginning of the value.

    '''
    ret = {'name': name,
           'result': True,
           'comment': '',
           'changes': {}}

    if not os.path.exists(name):
        ret['result'] = False
        ret['comment'] = "File or directory doesn't exist"
        return ret

    current_attrs = __salt__['xattr.list'](name)
    current_ids = current_attrs.keys()

    for attr in attributes:
        attr_id, attr_val = attr.split("=")
        attr_hex = attr_val.startswith("0x")

        if attr_hex:
            # Remove spaces and new lines so we can match these
            current_attrs[attr_id] = __salt__['xattr.read'](name, attr_id, hex=True).replace(" ", "").replace("\n", "")
            attr_val = attr_val[2:].replace(" ", "")

        if attr_id not in current_attrs:
            value_matches = False
        else:
            value_matches = ((current_attrs[attr_id] == attr_val) or
                             (attr_hex and current_attrs[attr_id] == attr_val))

        if attr_id in current_ids and value_matches:
            continue
        else:
            ret['changes'][attr_id] = attr_val
            __salt__['xattr.write'](name, attr_id, attr_val, attr_hex)

    if len(ret['changes'].keys()) == 0:
        ret['comment'] = 'All values existed correctly.'

    return ret


def delete(name, attributes):
    '''
    Make sure the given attributes are deleted from the file/directory

    name
        The path to the file/directory

    attributes
        The attributes that should be removed from the file/directory, this is accepted as
        an array.
    '''
    ret = {'name': name,
           'result': True,
           'comment': '',
           'changes': {}}

    if not os.path.exists(name):
        ret['result'] = False
        ret['comment'] = "File or directory doesn't exist"
        return ret

    current_attrs = __salt__['xattr.list'](name)
    current_ids = current_attrs.keys()

    for attr in attributes:

        if attr in current_ids:
            __salt__['xattr.delete'](name, attr)
            ret['changes'][attr] = 'delete'

    if len(ret['changes'].keys()) == 0:
        ret['comment'] = 'All attributes were already deleted.'

    return ret



# -*- coding: utf-8 -*-
'''
Store key/value pairs in a CSV file

.. versionadded:: 2016.11.0

Example configuration:

.. code-block:: yaml

    ext_pillar:
      - csv: /path/to/file.csv

    # or

    ext_pillar:
      - csv:
          path: /path/to/file.csv
          namespace: 'subkey'
          fieldnames:
          - col1
          - col2
          - col2

The first column must be minion IDs and the first row must be dictionary keys.
E.g.:

==========  =========   ======
id          role        env
==========  =========   ======
jerry       web         prod
stuart      web         stage
dave        web         qa
phil        db          prod
kevin       db          stage
mike        db          qa
==========  =========   ======

Will produce the following Pillar values for a minion named "jerry":

.. code-block:: python

    {
        'role': 'web',
        'env': 'prod',
    }
'''
from __future__ import absolute_import, print_function, unicode_literals
import csv

import salt.utils.files

__virtualname__ = 'csv'


def __virtual__():
    return __virtualname__


def ext_pillar(
        mid,
        pillar,
        path,
        idkey='id',
        namespace=None,
        fieldnames=None,
        restkey=None,
        restval=None,
        dialect='excel'):
    '''
    Read a CSV into Pillar

    :param str path: Absolute path to a CSV file.
    :param str idkey: (Optional) The column name of minion IDs.
    :param str namespace: (Optional) A pillar key to namespace the values under.
    :param list fieldnames: (Optional) if the first row of the CSV is not
        column names they may be specified here instead.
    '''
    with salt.utils.files.fopen(path, 'rb') as f:
        sheet = csv.DictReader(f, fieldnames,
                restkey=restkey, restval=restval, dialect=dialect)

        for row in sheet:
            if row[idkey] == mid:
                if namespace:
                    return {namespace: row}
                else:
                    return row

    return {}



# -*- coding: utf-8 -*-
'''
Common functions for managing mounts
'''

# Import python libs
from __future__ import absolute_import, unicode_literals
import logging
import os

# Import Salt libs
import salt.utils.files
import salt.utils.stringutils
import salt.utils.yaml
import salt.utils.versions

log = logging.getLogger(__name__)


def _read_file(path):
    '''
    Reads and returns the contents of a text file
    '''
    try:
        with salt.utils.files.fopen(path, 'rb') as contents:
            return salt.utils.yaml.safe_load(contents)
    except (OSError, IOError):
        return {}


def get_cache(opts):
    '''
    Return the mount cache file location.
    '''
    return os.path.join(opts['cachedir'], 'mounts')


def read_cache(opts):
    '''
    Write the mount cache file.
    '''
    cache_file = get_cache(opts)
    return _read_file(cache_file)


def write_cache(cache, opts):
    '''
    Write the mount cache file.
    '''
    cache_file = get_cache(opts)

    try:
        _cache = salt.utils.stringutils.to_bytes(
            salt.utils.yaml.safe_dump(cache)
        )
        with salt.utils.files.fopen(cache_file, 'wb+') as fp_:
            fp_.write(_cache)
        return True
    except (IOError, OSError):
        log.error('Failed to cache mounts',
                  exc_info_on_loglevel=logging.DEBUG)
        return False



# -*- coding: utf-8 -*-
'''
Initialize the engines system. This plugin system allows for
complex services to be encapsulated within the salt plugin environment
'''
# Import python libs
from __future__ import absolute_import, print_function, unicode_literals
import multiprocessing
import logging

# Import salt libs
import salt
import salt.loader
import salt.utils.platform
from salt.utils.process import SignalHandlingMultiprocessingProcess

log = logging.getLogger(__name__)


def start_engines(opts, proc_mgr, proxy=None):
    '''
    Fire up the configured engines!
    '''
    utils = salt.loader.utils(opts, proxy=proxy)
    if opts['__role'] == 'master':
        runners = salt.loader.runner(opts, utils=utils)
    else:
        runners = []
    funcs = salt.loader.minion_mods(opts, utils=utils, proxy=proxy)
    engines = salt.loader.engines(opts, funcs, runners, utils, proxy=proxy)

    engines_opt = opts.get('engines', [])
    if isinstance(engines_opt, dict):
        engines_opt = [{k: v} for k, v in engines_opt.items()]

    # Function references are not picklable. Windows needs to pickle when
    # spawning processes. On Windows, these will need to be recalculated
    # in the spawned child process.
    if salt.utils.platform.is_windows():
        runners = None
        utils = None
        funcs = None

    for engine in engines_opt:
        if isinstance(engine, dict):
            engine, engine_opts = next(iter(engine.items()))
        else:
            engine_opts = None
        fun = '{0}.start'.format(engine)
        if fun in engines:
            start_func = engines[fun]
            name = '{0}.Engine({1})'.format(__name__, start_func.__module__)
            log.info('Starting Engine %s', name)
            proc_mgr.add_process(
                    Engine,
                    args=(
                        opts,
                        fun,
                        engine_opts,
                        funcs,
                        runners,
                        proxy
                        ),
                    name=name
                    )


class Engine(SignalHandlingMultiprocessingProcess):
    '''
    Execute the given engine in a new process
    '''
    def __init__(self, opts, fun, config, funcs, runners, proxy, log_queue=None):
        '''
        Set up the process executor
        '''
        super(Engine, self).__init__(log_queue=log_queue)
        self.opts = opts
        self.config = config
        self.fun = fun
        self.funcs = funcs
        self.runners = runners
        self.proxy = proxy

    # __setstate__ and __getstate__ are only used on Windows.
    # We do this so that __init__ will be invoked on Windows in the child
    # process so that a register_after_fork() equivalent will work on Windows.
    def __setstate__(self, state):
        self._is_child = True
        self.__init__(
            state['opts'],
            state['fun'],
            state['config'],
            state['funcs'],
            state['runners'],
            state['proxy'],
            log_queue=state['log_queue']
        )

    def __getstate__(self):
        return {'opts': self.opts,
                'fun': self.fun,
                'config': self.config,
                'funcs': self.funcs,
                'runners': self.runners,
                'proxy': self.proxy,
                'log_queue': self.log_queue}

    def run(self):
        '''
        Run the master service!
        '''
        self.utils = salt.loader.utils(self.opts, proxy=self.proxy)
        if salt.utils.platform.is_windows():
            # Calculate function references since they can't be pickled.
            if self.opts['__role'] == 'master':
                self.runners = salt.loader.runner(self.opts, utils=self.utils)
            else:
                self.runners = []
            self.funcs = salt.loader.minion_mods(self.opts, utils=self.utils, proxy=self.proxy)

        self.engine = salt.loader.engines(self.opts,
                                          self.funcs,
                                          self.runners,
                                          self.utils,
                                          proxy=self.proxy)
        kwargs = self.config or {}
        try:
            self.engine[self.fun](**kwargs)
        except Exception as exc:
            log.critical(
                'Engine \'%s\' could not be started!',
                self.fun.split('.')[0], exc_info=True
            )



# -*- coding: utf-8 -*-
'''
Control of entries in SSH authorized_key files
==============================================

The information stored in a user's SSH authorized key file can be easily
controlled via the ssh_auth state. Defaults can be set by the enc, options,
and comment keys. These defaults can be overridden by including them in the
name.

Since the YAML specification limits the length of simple keys to 1024
characters, and since SSH keys are often longer than that, you may have
to use a YAML 'explicit key', as demonstrated in the second example below.

.. code-block:: yaml

    AAAAB3NzaC1kc3MAAACBAL0sQ9fJ5bYTEyY==:
      ssh_auth.present:
        - user: root
        - enc: ssh-dss

    ? AAAAB3NzaC1kc3MAAACBAL0sQ9fJ5bYTEyY==...
    :
      ssh_auth.present:
        - user: root
        - enc: ssh-dss

    thatch:
      ssh_auth.present:
        - user: root
        - source: salt://ssh_keys/thatch.id_rsa.pub
        - config: '%h/.ssh/authorized_keys'

    sshkeys:
      ssh_auth.present:
        - user: root
        - enc: ssh-rsa
        - options:
          - option1="value1"
          - option2="value2 flag2"
        - comment: myuser
        - names:
          - AAAAB3NzaC1kc3MAAACBAL0sQ9fJ5bYTEyY==
          - ssh-dss AAAAB3NzaCL0sQ9fJ5bYTEyY== user@domain
          - option3="value3" ssh-dss AAAAB3NzaC1kcQ9J5bYTEyY== other@testdomain
          - AAAAB3NzaC1kcQ9fJFF435bYTEyY== newcomment
'''

# Import python libs
from __future__ import absolute_import, unicode_literals, print_function
import re
import sys

# Import 3rd-party libs
from salt.ext import six


def _present_test(user, name, enc, comment, options, source, config, fingerprint_hash_type):
    '''
    Run checks for "present"
    '''
    result = None
    if source:
        keys = __salt__['ssh.check_key_file'](
                user,
                source,
                config,
                saltenv=__env__,
                fingerprint_hash_type=fingerprint_hash_type)
        if keys:
            comment = ''
            for key, status in six.iteritems(keys):
                if status == 'exists':
                    continue
                comment += 'Set to {0}: {1}\n'.format(status, key)
            if comment:
                return result, comment
        err = sys.modules[
            __salt__['test.ping'].__module__
        ].__context__.pop('ssh_auth.error', None)
        if err:
            return False, err
        else:
            return (
                True,
                'All host keys in file {0} are already present'.format(source)
            )
    else:
        # check if this is of form {options} {enc} {key} {comment}
        sshre = re.compile(r'^(.*?)\s?((?:ssh\-|ecds)[\w-]+\s.+)$')
        fullkey = sshre.search(name)
        # if it is {key} [comment]
        if not fullkey:
            key_and_comment = name.split()
            name = key_and_comment[0]
            if len(key_and_comment) == 2:
                comment = key_and_comment[1]
        else:
            # if there are options, set them
            if fullkey.group(1):
                options = fullkey.group(1).split(',')
            # key is of format: {enc} {key} [comment]
            comps = fullkey.group(2).split()
            enc = comps[0]
            name = comps[1]
            if len(comps) == 3:
                comment = comps[2]

    check = __salt__['ssh.check_key'](
            user,
            name,
            enc,
            comment,
            options,
            config=config,
            fingerprint_hash_type=fingerprint_hash_type)
    if check == 'update':
        comment = (
                'Key {0} for user {1} is set to be updated'
                ).format(name, user)
    elif check == 'add':
        comment = (
                'Key {0} for user {1} is set to be added'
                ).format(name, user)
    elif check == 'exists':
        result = True
        comment = ('The authorized host key {0} is already present '
                          'for user {1}'.format(name, user))

    return result, comment


def _absent_test(user, name, enc, comment, options, source, config, fingerprint_hash_type):
    '''
    Run checks for "absent"
    '''
    result = None
    if source:
        keys = __salt__['ssh.check_key_file'](
                user,
                source,
                config,
                saltenv=__env__,
                fingerprint_hash_type=fingerprint_hash_type)
        if keys:
            comment = ''
            for key, status in list(keys.items()):
                if status == 'add':
                    continue
                comment += 'Set to remove: {0}\n'.format(key)
            if comment:
                return result, comment
        err = sys.modules[
            __salt__['test.ping'].__module__
        ].__context__.pop('ssh_auth.error', None)
        if err:
            return False, err
        else:
            return (
                True,
                'All host keys in file {0} are already absent'.format(source)
            )
    else:
        # check if this is of form {options} {enc} {key} {comment}
        sshre = re.compile(r'^(.*?)\s?((?:ssh\-|ecds)[\w-]+\s.+)$')
        fullkey = sshre.search(name)
        # if it is {key} [comment]
        if not fullkey:
            key_and_comment = name.split()
            name = key_and_comment[0]
            if len(key_and_comment) == 2:
                comment = key_and_comment[1]
        else:
            # if there are options, set them
            if fullkey.group(1):
                options = fullkey.group(1).split(',')
            # key is of format: {enc} {key} [comment]
            comps = fullkey.group(2).split()
            enc = comps[0]
            name = comps[1]
            if len(comps) == 3:
                comment = comps[2]

    check = __salt__['ssh.check_key'](
            user,
            name,
            enc,
            comment,
            options,
            config=config,
            fingerprint_hash_type=fingerprint_hash_type)
    if check == 'update' or check == 'exists':
        comment = ('Key {0} for user {1} is set for removal').format(name, user)
    else:
        comment = ('Key is already absent')
        result = True

    return result, comment


def present(
        name,
        user,
        enc='ssh-rsa',
        comment='',
        source='',
        options=None,
        config='.ssh/authorized_keys',
        fingerprint_hash_type=None,
        **kwargs):
    '''
    Verifies that the specified SSH key is present for the specified user

    name
        The SSH key to manage

    user
        The user who owns the SSH authorized keys file to modify

    enc
        Defines what type of key is being used; can be ed25519, ecdsa, ssh-rsa
        or ssh-dss

    comment
        The comment to be placed with the SSH public key

    source
        The source file for the key(s). Can contain any number of public keys,
        in standard "authorized_keys" format. If this is set, comment and enc
        will be ignored.

    .. note::
        The source file must contain keys in the format ``<enc> <key>
        <comment>``. If you have generated a keypair using PuTTYgen, then you
        will need to do the following to retrieve an OpenSSH-compatible public
        key.

        1. In PuTTYgen, click ``Load``, and select the *private* key file (not
           the public key), and click ``Open``.
        2. Copy the public key from the box labeled ``Public key for pasting
           into OpenSSH authorized_keys file``.
        3. Paste it into a new file.

    options
        The options passed to the key, pass a list object

    config
        The location of the authorized keys file relative to the user's home
        directory, defaults to ".ssh/authorized_keys". Token expansion %u and
        %h for username and home path supported.

    fingerprint_hash_type
        The public key fingerprint hash type that the public key fingerprint
        was originally hashed with. This defaults to ``md5`` if not specified.

        .. versionadded:: 2016.11.7

        .. note::

            The default value of the ``fingerprint_hash_type`` will change to
            ``sha256`` in Salt 2017.7.0.
    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': ''}

    if source == '':
        # check if this is of form {options} {enc} {key} {comment}
        sshre = re.compile(r'^(.*?)\s?((?:ssh\-|ecds)[\w-]+\s.+)$')
        fullkey = sshre.search(name)
        # if it is {key} [comment]
        if not fullkey:
            key_and_comment = name.split(None, 1)
            name = key_and_comment[0]
            if len(key_and_comment) == 2:
                comment = key_and_comment[1]
        else:
            # if there are options, set them
            if fullkey.group(1):
                options = fullkey.group(1).split(',')
            # key is of format: {enc} {key} [comment]
            comps = fullkey.group(2).split(None, 2)
            enc = comps[0]
            name = comps[1]
            if len(comps) == 3:
                comment = comps[2]

    if __opts__['test']:
        ret['result'], ret['comment'] = _present_test(
                user,
                name,
                enc,
                comment,
                options or [],
                source,
                config,
                fingerprint_hash_type)
        return ret

    # Get only the path to the file without env referrences to check if exists
    if source != '':
        source_path = __salt__['cp.get_url'](
                        source,
                        None,
                        saltenv=__env__)

    if source != '' and not source_path:
        data = 'no key'
    elif source != '' and source_path:
        key = __salt__['cp.get_file_str'](
                source,
                saltenv=__env__)
        filehasoptions = False
        # check if this is of form {options} {enc} {key} {comment}
        sshre = re.compile(r'^(ssh\-|ecds).*')
        key = key.rstrip().split('\n')
        for keyline in key:
            filehasoptions = sshre.match(keyline)
            if not filehasoptions:
                data = __salt__['ssh.set_auth_key_from_file'](
                        user,
                        source,
                        config=config,
                        saltenv=__env__,
                        fingerprint_hash_type=fingerprint_hash_type)
            else:
                # Split keyline to get key und comment
                keyline = keyline.split(' ')
                key_type = keyline[0]
                key_value = keyline[1]
                key_comment = keyline[2] if len(keyline) > 2 else ''
                data = __salt__['ssh.set_auth_key'](
                        user,
                        key_value,
                        enc=key_type,
                        comment=key_comment,
                        options=options or [],
                        config=config,
                        fingerprint_hash_type=fingerprint_hash_type)
    else:
        data = __salt__['ssh.set_auth_key'](
                user,
                name,
                enc=enc,
                comment=comment,
                options=options or [],
                config=config,
                fingerprint_hash_type=fingerprint_hash_type)

    if data == 'replace':
        ret['changes'][name] = 'Updated'
        ret['comment'] = ('The authorized host key {0} for user {1} was '
                          'updated'.format(name, user))
        return ret
    elif data == 'no change':
        ret['comment'] = ('The authorized host key {0} is already present '
                          'for user {1}'.format(name, user))
    elif data == 'new':
        ret['changes'][name] = 'New'
        ret['comment'] = ('The authorized host key {0} for user {1} was added'
                          .format(name, user))
    elif data == 'no key':
        ret['result'] = False
        ret['comment'] = ('Failed to add the ssh key. Source file {0} is '
                          'missing'.format(source))
    elif data == 'fail':
        ret['result'] = False
        err = sys.modules[
            __salt__['test.ping'].__module__
        ].__context__.pop('ssh_auth.error', None)
        if err:
            ret['comment'] = err
        else:
            ret['comment'] = ('Failed to add the ssh key. Is the home '
                              'directory available, and/or does the key file '
                              'exist?')
    elif data == 'invalid' or data == 'Invalid public key':
        ret['result'] = False
        ret['comment'] = 'Invalid public ssh key, most likely has spaces or invalid syntax'

    return ret


def absent(name,
           user,
           enc='ssh-rsa',
           comment='',
           source='',
           options=None,
           config='.ssh/authorized_keys',
           fingerprint_hash_type=None):
    '''
    Verifies that the specified SSH key is absent

    name
        The SSH key to manage

    user
        The user who owns the SSH authorized keys file to modify

    enc
        Defines what type of key is being used; can be ed25519, ecdsa, ssh-rsa
        or ssh-dss

    comment
        The comment to be placed with the SSH public key

    options
        The options passed to the key, pass a list object

    source
        The source file for the key(s). Can contain any number of public keys,
        in standard "authorized_keys" format. If this is set, comment, enc and
        options will be ignored.

        .. versionadded:: 2015.8.0

    config
        The location of the authorized keys file relative to the user's home
        directory, defaults to ".ssh/authorized_keys". Token expansion %u and
        %h for username and home path supported.

    fingerprint_hash_type
        The public key fingerprint hash type that the public key fingerprint
        was originally hashed with. This defaults to ``md5`` if not specified.

        .. versionadded:: 2016.11.7

        .. note::

            The default value of the ``fingerprint_hash_type`` will change to
            ``sha256`` in Salt 2017.7.0.

    '''
    ret = {'name': name,
           'changes': {},
           'result': True,
           'comment': ''}

    if __opts__['test']:
        ret['result'], ret['comment'] = _absent_test(
                user,
                name,
                enc,
                comment,
                options or [],
                source,
                config,
                fingerprint_hash_type)
        return ret

    # Extract Key from file if source is present
    if source != '':
        key = __salt__['cp.get_file_str'](
                source,
                saltenv=__env__)
        filehasoptions = False
        # check if this is of form {options} {enc} {key} {comment}
        sshre = re.compile(r'^(ssh\-|ecds).*')
        key = key.rstrip().split('\n')
        for keyline in key:
            filehasoptions = sshre.match(keyline)
            if not filehasoptions:
                ret['comment'] = __salt__['ssh.rm_auth_key_from_file'](user,
                                                                       source,
                                                                       config,
                                                                       saltenv=__env__,
                                                                       fingerprint_hash_type=fingerprint_hash_type)
            else:
                # Split keyline to get key
                keyline = keyline.split(' ')
                ret['comment'] = __salt__['ssh.rm_auth_key'](user,
                                                             keyline[1],
                                                             config=config,
                                                             fingerprint_hash_type=fingerprint_hash_type)
    else:
        # Get just the key
        sshre = re.compile(r'^(.*?)\s?((?:ssh\-|ecds)[\w-]+\s.+)$')
        fullkey = sshre.search(name)
        # if it is {key} [comment]
        if not fullkey:
            key_and_comment = name.split(None, 1)
            name = key_and_comment[0]
            if len(key_and_comment) == 2:
                comment = key_and_comment[1]
        else:
            # if there are options, set them
            if fullkey.group(1):
                options = fullkey.group(1).split(',')
            # key is of format: {enc} {key} [comment]
            comps = fullkey.group(2).split()
            enc = comps[0]
            name = comps[1]
            if len(comps) == 3:
                comment = comps[2]
        ret['comment'] = __salt__['ssh.rm_auth_key'](user,
                                                     name,
                                                     config=config,
                                                     fingerprint_hash_type=fingerprint_hash_type)

    if ret['comment'] == 'User authorized keys file not present':
        ret['result'] = False
        return ret
    elif ret['comment'] == 'Key removed':
        ret['changes'][name] = 'Removed'

    return ret



# -*- coding: utf-8 -*-
'''
Use pycrypto to generate random passwords on the fly.
'''

# Import python libraries
from __future__ import absolute_import, print_function, unicode_literals
import logging
import re
import string
import random

# Import 3rd-party libs
try:
    try:
        from M2Crypto.Rand import rand_bytes as get_random_bytes
    except ImportError:
        try:
            from Cryptodome.Random import get_random_bytes  # pylint: disable=E0611
        except ImportError:
            from Crypto.Random import get_random_bytes  # pylint: disable=E0611
    HAS_RANDOM = True
except ImportError:
    HAS_RANDOM = False

try:
    # Windows does not have the crypt module
    import crypt
    HAS_CRYPT = True
except ImportError:
    HAS_CRYPT = False

# Import salt libs
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, SaltInvocationError
from salt.ext import six

log = logging.getLogger(__name__)


def secure_password(length=20, use_random=True):
    '''
    Generate a secure password.
    '''
    try:
        length = int(length)
        pw = ''
        while len(pw) < length:
            if HAS_RANDOM and use_random:
                while True:
                    try:
                        char = salt.utils.stringutils.to_str(get_random_bytes(1))
                        break
                    except UnicodeDecodeError:
                        continue
                pw += re.sub(
                    salt.utils.stringutils.to_str(r'\W'),
                    str(),  # future lint: disable=blacklisted-function
                    char
                )
            else:
                pw += random.SystemRandom().choice(string.ascii_letters + string.digits)
        return pw
    except Exception as exc:
        log.exception('Failed to generate secure passsword')
        raise CommandExecutionError(six.text_type(exc))


def gen_hash(crypt_salt=None, password=None, algorithm='sha512'):
    '''
    Generate /etc/shadow hash
    '''
    if not HAS_CRYPT:
        raise SaltInvocationError('No crypt module for windows')

    hash_algorithms = dict(
        md5='$1$', blowfish='$2a$', sha256='$5$', sha512='$6$'
    )
    if algorithm not in hash_algorithms:
        raise SaltInvocationError(
            'Algorithm \'{0}\' is not supported'.format(algorithm)
        )

    if password is None:
        password = secure_password()

    if crypt_salt is None:
        crypt_salt = secure_password(8)

    crypt_salt = hash_algorithms[algorithm] + crypt_salt

    return crypt.crypt(password, crypt_salt)
